package com.vanheusden.pfa;

import java.util.*;

abstract class ChessObject {
	protected int x, y;
	protected final PlayerColor color;
	protected int nMoves; /* used for rokkade/en passant */

	protected boolean isPinned = false, isKingPinned = false;

	/* statistics */
        protected int thisCanBeCaughtBy = 0, canDoNMoves = 0;
        protected boolean generatedListOfMoves;

	/* move list generation */
	protected Board board;
	protected List<Move> moves;

	void addMove(final ChessObject targetField, final int fx, final int fy, final int tx, final int ty, final MoveType moveType) {
		if (targetField != null)
			targetField.addCatcher();

		moves.add(new Move(fx, fy, tx, ty, moveType));

		canDoNMoves++;
	}

	void addMove(final int fx, final int fy, final int tx, final int ty) {
		moves.add(new Move(fx, fy, tx, ty));

		canDoNMoves++;
	}

	void addMove(final ChessObject targetField, final int fx, final int fy, final int tx, final int ty) {
		if (targetField != null)
			targetField.addCatcher();

		moves.add(new Move(fx, fy, tx, ty));

		canDoNMoves++;
	}

	/* checks if destination field is not occupied by a piece of myself
	 * returns 'true' when a stop condition is encountered:
	 * - beyond board
	 * - board-location with object
	 */
	boolean conditionalAddMove(final int fx, final int fy, final int tx, final int ty) {
		final ChessObject destField = board.getAt(tx, ty);

		if (destField == null) {
			addMove(destField, fx, fy, tx, ty);

			return false;
		}

		if (destField.getColor() != color)
			addMove(destField, fx, fy, tx, ty);

		return true;
	}

	void getMovesPath(final int deltaX, final int deltaY) {
		ChessObject pinnedPiece = null;
		int cx = x + deltaX, cy = y + deltaY;

		while(cx >= 0 && cy >= 0 && cx <= 7 && cy <= 7) {
			final ChessObject destField = board.getAt(cx, cy);

			if (destField != null) {
				if (pinnedPiece == null) {
					// if this piece is of the opponent, remember this
					// move and continue searching for another of the
					// opponent. then if another is found, this piece
					// (destField) is pinned.
					if (destField.getColor() != color) {
						addMove(destField, x, y, cx, cy);

						pinnedPiece = destField;
					}
					// if the piece we encounter is one of mine, then
					// stop searching
					else {
						break;
					}
				}
				else {
					if (destField.getColor() != color) {
						pinnedPiece.setPinned();

						if (destField instanceof ChessObjectKing)
							pinnedPiece.setKingPinned();
					}

					break;
				}
			}

			if (pinnedPiece == null)
				addMove(x, y, cx, cy);

			cx += deltaX;
			cy += deltaY;
		}
	}

	ChessObject(final int x, final int y, final PlayerColor color) {
		this.x = x;
		this.y = y;
		this.color = color;
		this.nMoves = 0;
	}

	ChessObject(final int x, final int y, final PlayerColor color, final int nMoves) {
		assert nMoves >= 0;
		this.x = x;
		this.y = y;
		this.color = color;
		this.nMoves = nMoves;
	}

	////////////////////////////////////////////////////////////////
	// SHOULD ONLY BE CALLED BY SCENE OBJECT!!!
	void doMove(final int x, final int y) {
		assert x >= 0 && y >= 0 && x < 8 && y < 8;

		this.x = x;
		this.y = y;

		nMoves++;
	}

	void doMove(final Move move) {
		assert move.getFX() == x && move.getFY() == y;

		doMove(move.getTX(), move.getTY());
	}

	// SHOULD ONLY BE CALLED BY FEN-PARSER!!!
	void resetNMoves() {
		nMoves = 0;
	}
	void setNMoves() {
		nMoves = 696;
	}
	////////////////////////////////////////////////////////////////

	public String toString() {
		return "type: " + this.getClass().getName() +
			", x/y: " + x + "/" + y +
			", color: " + color +
			", did nMOves: " + nMoves +
			", list of moves is " + (generatedListOfMoves ? "" : "not ") + "generated" +
			", can be caught by " + thisCanBeCaughtBy;
	}

	ChessObject duplicate() {
		return null;
	}

	ChessObjectType getType() {
		return null;
	}

	String getTypeCharacter(final boolean uppercase) {
		return null;
	}

	PlayerColor getColor() {
		return color;
	}

	int getX() {
		return x;
	}

	int getY() {
		return y;
	}

	boolean isFirstMove() {
		return nMoves == 0;
	}

	int getNMoves() {
		return nMoves;
	}

	void getMoves(final Scene scene, final List<Move> moves, final ChessObject king, final Move lastMove) {
		assert false;
	}

	ChessObject promote(final ChessObjectType newType) {
		assert false;
		return null;
	}

	void addCatcher() {
		thisCanBeCaughtBy++;

		assert thisCanBeCaughtBy <= 16;
	}

	int getNCatchers() {
		return thisCanBeCaughtBy;
	}

	boolean getPinned() {
		return isPinned;
	}

	void setPinned() {
		isPinned = true;
	}

	boolean getKingPinned() {
		return isKingPinned;
	}

	void setKingPinned() {
		isKingPinned = true;
	}

	int getCanDoNMoves() {
		return canDoNMoves;
	}

	public static ChessObject newChessObject(final ChessObjectType type, final int x, final int y, final PlayerColor color) {
		switch(type) {
			case PAWN:
				return new ChessObjectPawn(x, y, color);
			case ROOK:
				return new ChessObjectRook(x, y, color);
			case BISHOP:
				return new ChessObjectBishop(x, y, color);
			case KNIGHT:
				return new ChessObjectKnight(x, y, color);
			case KING:
				return new ChessObjectKing(x, y, color);
			case QUEEN:
				return new ChessObjectQueen(x, y, color);
		}

		assert false;

		return null;
	}
}
