package com.vanheusden.pfa;

import java.io.*;
import java.util.*;

public class DeepBrutePos {
	static String logFile;
	static final int defaultTransTabSize = 512 * 1024;
	static final int defaultThreadsPerSlot = 8;
	static final int defaultMaxSearchDuration = 30;
	static final int defaultMaxExtensions = 5;

	public void showHelp() {
		System.out.println("--io-mode [xboard|uci|ucisocket|console|consoleansi|test|perft|testcases|tournament|elotester]");
		System.out.println("\txboard:\t\tcompatible with xboard/winboard");
		System.out.println("\tuci:\t\tcompatible with UCI (e.g. Arena)");
		System.out.println("\tucisocket:\tUCI, listening on a TCP network socket. see --port");
		System.out.println("\tconsole:\tplain ascii output usable on dumb terminals");
		System.out.println("\tconsoleansi:\tconsole alike mode using ANSI escape codes for colors");
		System.out.println("\ttest:\t\truns a game playing against itself, --epd-position can be used");
		System.out.println("\ttestansi:\tlike 'test', see consoleansi, --epd-position can be used");
		System.out.println("\tperft:\t\tthis is a mode usable to test if the move-generator is implemented correctly");
		System.out.println("\ttestcases:\truns a couple of tests to detect any bugs in the code");
		System.out.println("\ttournament:\truns a tournament between 2 or more xboard/uci chess engines. see --drone, set a limit on the number of moves using --move-n-limit");
		System.out.println("\telotester:\ttest from chessmaniac.com to test the strength (in ELO points) of one or more chess engines. see --drone");
		System.out.println("--max-depth");
		System.out.println("\tsearchdepth (default is 0 (plies) which means no limit, minimum is 1)");
		System.out.println("--random");
		System.out.println("\tenables randomization");
		System.out.println("--max-n-cores");
		System.out.println("\tmaximum number of threads, default is system limit");
		System.out.println("--max-search-duration");
		System.out.println("\thow long to think, in seconds (default: " + defaultMaxSearchDuration + ")");
		System.out.println("--trans-tab-size");
		System.out.println("\tsize of transposition table, in entries. each entry is +/- 90 bytes (default: " + defaultTransTabSize + " elements)");
		System.out.println("--trans-tab-file");
		System.out.println("\tfile to serialize the transposition table to");
		System.out.println("--trans-tab-sync-interval");
		System.out.println("\tsync transposition table every x seconds (default: disabled)");
		System.out.println("--brain x");
		System.out.println("\tx=random: select random mover");
		System.out.println("\tx=firstmove: always select first move from move list");
		System.out.println("--ponder");
		System.out.println("\talso think in opponent time");
		System.out.println("--max-threads-per-slot (default: " + defaultThreadsPerSlot + ")");
		System.out.println("--extensions (default: " + defaultMaxExtensions + ")");
		System.out.println("--disable-quiescence");
		System.out.println("--book-best-move");
		System.out.println("\tbest move book database to use: url,user,password,bookname with url: jdbc:mysql://HOST:PORT/DATABASE");
		System.out.println("--book-wdl-static");
		System.out.println("\tWDL book database to use: url,user,password,bookname with url: jdbc:mysql://HOST:PORT/DATABASE");
		System.out.println("--ignore-protocol-overrides");
		System.out.println("\tdo not allow UCI/XBoard to set ponder/cores settings");
		System.out.println("--epd-position");
		System.out.println("\tsetup of board");
		System.out.println("--port");
		System.out.println("\tfor ucisocket: TCP port to listen on");
		System.out.println("--loop");
		System.out.println("\tfor ucisocket: when the client terminates, wait for next client instead of terminating");
		System.out.println("--profiler");
		System.out.println("--drone mode,host,port");
		System.out.println("\tswitch to \"puppet master\"-mode and connect to a TCP socket with an UCI-listener behind it / also used for \"--io-mode tournament\"");
		System.out.println("--pgnfile file.pgn");
		System.out.println("--logfile file.log  (human readable)");
		System.out.println("--debug enable debug mode");
		System.out.println("--version-add string to add to name, usefull for PGN output");
	}

	public static int getLineNumber() {
		return Thread.currentThread().getStackTrace()[2].getLineNumber();
	}

	enum BrainType { REGULAR, RANDOM, FIRSTMOVE };

	public static void main(String [] args) throws Exception {
		try {
			DeepBrutePos P = new DeepBrutePos();
			String strIOMode = "uci";
			IO io = null;
			String pgnFile = null;
			int depth = 0, maxExtension = defaultMaxExtensions, maxNProcs = 0;
			double maxSearchDuration = defaultMaxSearchDuration;
			boolean maxSearchDurationSet = false;
			boolean random = false;
			String transTabFile = null;
			int transTabSize = defaultTransTabSize, threadsPerSlot = defaultThreadsPerSlot;
			int syncTPTInterval = -1;
			String ioEPDPosTest = null;
			boolean allowQueiescence = true;
			Profiler profiler = null;
			int port = 55223;
			boolean socketLoop = false;
			List<Drone> drones = new ArrayList<Drone>();
			int maxNMoves = -1;
			BrainType bt = BrainType.REGULAR;
			List<Book> books = new ArrayList<Book>();
			boolean ponder = false;
			boolean ignore_protocol_overrides = false;
			String brainDump = null;
			List<String> beepers = new ArrayList<String>();

			for(int argIndex=0; argIndex<args.length; argIndex++) {
				if (args[argIndex].equals("--help")) {
					P.showHelp();
					System.exit(0);
				}
				else if (args[argIndex].equals("--io-mode")) {
					strIOMode = args[++argIndex];
				}
				else if (args[argIndex].equals("--profiler")) {
					profiler = new Profiler(true, 5);
				}
				else if (args[argIndex].equals("--opening-book-db")) {
					books.add(new BookBestMove(args[++argIndex]));
				}
				else if (args[argIndex].equals("--book-wdl-static")) {
					books.add(new BookWDLStatic(args[++argIndex]));
				}
				else if (args[argIndex].equals("--logfile")) {
					logFile = args[++argIndex];
					Statics.setLogFile(logFile);
				}
				else if (args[argIndex].equals("--version-add")) {
					Version.setAdd(args[++argIndex]);
				}
				else if (args[argIndex].equals("--epd-position")) {
					ioEPDPosTest = args[++argIndex];
				}
				else if (args[argIndex].equals("--pgnfile")) {
					pgnFile = args[++argIndex];
				}
				else if (args[argIndex].equals("--beeper")) {
					beepers.add(args[++argIndex]);
				}
				else if (args[argIndex].equals("--ponder")) {
					ponder = true;
				}
				else if (args[argIndex].equals("--ignore-protocol-overrides")) {
					ignore_protocol_overrides = true;
				}
				else if (args[argIndex].equals("--brain")) {
					String which = args[++argIndex];

					if (which.equals("normal") || which.equals("regular")) {
					}
					else if (which.equals("random")) {
						bt = BrainType.RANDOM;

						Version.setAdd("random-mover");
					}
					else if (which.equals("firstmove")) {
						bt = BrainType.FIRSTMOVE;

						Version.setAdd("firstmove");
					}
				}
				else if (args[argIndex].equals("--move-n-limit") || args[argIndex].equals("--max-n-moves")) {
					maxNMoves = Integer.valueOf(args[++argIndex]);
				}
				else if (args[argIndex].equals("--puppet") || args[argIndex].equals("--drone")) {
					String pars = args[++argIndex];

					String [] parts = pars.split(",");

					if (parts[0].equals("tcp-uci") == false && parts[0].equals("tcp-xboard") == false && parts[0].equals("exe-uci") == false && parts[0].equals("exe-xboard") == false) {
						System.err.println("--puppet requires a parameter with format: type,host/exe[,port] with type being either tcp-xboard, tcp-uci, exe-xboard or exe-uci");
						System.exit(1);
					}

					Map<String, String> droneParameters = new HashMap<String, String>();
					for(int pindex=3; pindex<parts.length; pindex++) {
						String [] fields = parts[pindex].split("=");

						droneParameters.put(fields[0], fields[1]);
					}

					String protocol = parts[0].substring(4);
					String type = parts[0].substring(0, 3);

					Statics.log("Adding drone: " + pars);

					if (protocol.equals("uci")) {
						if (type.equals("tcp"))
							drones.add(new UciDrone(parts[1], Integer.valueOf(parts[2]), true, droneParameters));
						else
							drones.add(new UciDrone(parts[1], null, true, droneParameters));
					}
					else {
						if (type.equals("tcp"))
							drones.add(new XboardDrone(parts[1], Integer.valueOf(parts[2]), true, droneParameters));
						else
							drones.add(new XboardDrone(parts[1], null, true, droneParameters));
					}

					Statics.log("Drone added");
				}
				else if (args[argIndex].equals("--loop")) {
					socketLoop = true;
				}
				else if (args[argIndex].equals("--random")) {
					random = true;
				}
				else if (args[argIndex].equals("--disable-quiescence")) {
					allowQueiescence = false;
				}
				else if (args[argIndex].equals("--extensions") || args[argIndex].equals("--max-extensions")) {
					maxExtension = Integer.valueOf(args[++argIndex]);
				}
				else if (args[argIndex].compareTo("--debug") == 0) {
					Statics.setDebug(true);
				}
				else if (args[argIndex].compareTo("--max-threads-per-slot") == 0) {
					threadsPerSlot = Integer.valueOf(args[++argIndex]);
				}
				else if (args[argIndex].equals("--trans-tab-file")) {
					transTabFile = args[++argIndex];
				}
				else if (args[argIndex].equals("--brain-dump")) {
					brainDump = args[++argIndex];
				}
				else if (args[argIndex].equals("--trans-tab-sync-interval")) {
					syncTPTInterval = (int)(Double.valueOf(args[++argIndex]) * 1000);
				}
				else if (args[argIndex].equals("--trans-tab-size")) {
					transTabSize = Integer.valueOf(args[++argIndex]);

					if (transTabSize < 0) {
						System.err.println("--trans-tab-size value must be >= 0 and < 2^31");
						System.exit(1);
					}
				}
				else if (args[argIndex].equals("--depth") || args[argIndex].equals("--max-depth") || args[argIndex].equals("--max-search-depth")) {
					depth = Integer.valueOf(args[++argIndex]);
					if (depth != 0 && depth < 1) {
						System.err.println("Invalid value for --depth, must be either 0 or >= 1");
						System.exit(1);
					}
				}
				else if (args[argIndex].equals("--max-search-duration") || args[argIndex].equals("--max-duration")) {
					maxSearchDuration = Double.valueOf(args[++argIndex]);
					if (depth < 0) {
						System.err.println("Invalid value for --depth, must be >= 0");
						System.exit(1);
					}
					maxSearchDurationSet = true;
				}
				else if (args[argIndex].equals("--max-n-procs") || args[argIndex].equals("--max-n-cores")) {
					maxNProcs = Integer.valueOf(args[++argIndex]);
					if (maxNProcs < 1) {
						System.err.println("Invalid value for --max-n-procs, must be > 0");
						System.exit(1);
					}
				}
				else if (args[argIndex].equals("--random")) {
				}
				else if (args[argIndex].equals("--verbose")) {
					Statics.increaseVerbosity();
				}
				else {
					System.err.println(args[argIndex] + " is not a recognised commandline switch.");
					System.err.println("");

					P.showHelp();

					System.exit(1);
				}
			}

			Statics.updateStartupLine();

			if (transTabSize == 0 && ponder) {
				final String msg = "Disabling ponder: no transposition- (hash-) table";

				System.err.println(msg);
				Statics.log(msg);
			}

			// banner
			if (!strIOMode.equals("uci") && !strIOMode.equals("epdcount"))
				System.out.println(Statics.startupLine);

			Brain brain = null;

			if (strIOMode.equals("tournament") || strIOMode.equals("tournamentsimple") || strIOMode.equals("elotester")) {
				// these do not require a brain
			}
			else if (drones.size() > 0)
				brain = new BrainPuppetMaster(books, drones, maxNProcs, ponder, brainDump);
			else if (bt == BrainType.REGULAR)
				brain = new Brain(transTabFile, transTabSize, syncTPTInterval, maxNProcs, threadsPerSlot, books, allowQueiescence);
			else if (bt == BrainType.RANDOM)
				brain = new BrainRandom();
			else if (bt == BrainType.FIRSTMOVE)
				brain = new BrainFirstMove();

			if ((strIOMode.equals("uci") || strIOMode.equals("xboard")) && maxSearchDurationSet == false)
				maxSearchDuration = -1;

			// instantiate modes
			if (strIOMode.equals("uci")) {
				io = new IO_UCI(pgnFile, depth, maxExtension, random, maxSearchDuration, brain, ponder, ignore_protocol_overrides, beepers);
			}
			else if (strIOMode.equals("console")) {
				io = new IO_Console(pgnFile, depth, maxExtension, random, maxSearchDuration, ioEPDPosTest, brain, ponder, beepers);
			}
			else if (strIOMode.equals("consoleansi")) {
				io = new IO_ConsoleANSI(pgnFile, depth, maxExtension, random, maxSearchDuration, ioEPDPosTest, brain, ponder, beepers);
			}
			else if (strIOMode.equals("xboard")) {
				io = new IO_XBoard(pgnFile, depth, maxExtension, random, maxSearchDuration, brain, ponder, ignore_protocol_overrides, beepers);
			}
			else if (strIOMode.equals("test")) {
				io = new IO_Test(pgnFile, depth, maxExtension, random, maxSearchDuration, brain, ioEPDPosTest, ponder, beepers);
			}
			else if (strIOMode.equals("epdcount")) {
				io = new IO_EPDCount(pgnFile, depth, maxExtension, random, maxSearchDuration, brain, ioEPDPosTest, ponder, beepers);
			}
			else if (strIOMode.equals("testansi")) {
				io = new IO_TestANSI(pgnFile, depth, maxExtension, random, maxSearchDuration, brain, ioEPDPosTest, ponder, beepers);
			}
			else if (strIOMode.equals("testcases")) {
				io = new IO_Testcases(depth, maxExtension, random, maxSearchDuration, brain, ponder, beepers);
			}
			else if (strIOMode.equals("ucisocket")) {
				io = new IO_UCI_socket(pgnFile, depth, maxExtension, random, maxSearchDuration, port, brain, ponder, ignore_protocol_overrides, beepers);
				System.out.println("Listening on port " + port);
			}
			else if (strIOMode.equals("perft")) {
				for(int curDepth = 1; curDepth <= depth; curDepth++) {
					Perft p = new Perft(Statics.getVerbosity() > 0);

					long startedAt = System.currentTimeMillis();
					long n = p.perft(ioEPDPosTest, curDepth, maxNProcs);
					System.out.print("Nodes at " + curDepth + ": " + n);
					long finishedAt = System.currentTimeMillis();
					double took = (finishedAt - startedAt) / 1000.0;
					System.out.println(" (took: " + took + "s, " + n / took + "nodes/s)");
				}
			}
			else if (strIOMode.equals("tournament")) {
				io = new IO_Tournament(pgnFile, depth, maxExtension, random, maxSearchDuration, ioEPDPosTest, drones, maxNMoves, ponder, beepers);
			}
			else if (strIOMode.equals("tournamentsimple")) {
				io = new IO_TournamentSimple(pgnFile, depth, maxExtension, random, maxSearchDuration, ioEPDPosTest, drones, maxNMoves, ponder, beepers);
			}
			else if (strIOMode.equals("elotester")) {
				io = new IO_EloTester(depth, maxExtension, random, maxSearchDuration, drones, ponder, beepers);
			}
			else {
				System.out.println("IO-mode '" + strIOMode + "' is not recognized.");
				P.showHelp();
			}

			Statics.log("Go!");

			if (io != null) {
				do {
					io.loop();
				}
				while(socketLoop);
			}

			if (brain != null)
				brain.destructor();

			if (profiler != null) {
				profiler.stopThread();

				profiler.dump();
			}
		}
		// FIXME catch(Exception e) {
		catch(InterruptedException ie) {
		}
		catch(Exception e) {
			System.out.println("\n----------------------------------------------------------------------------\n");
			System.out.println("An exception occured. Please report the following to folkert@vanheusden.com\n");
			if (logFile != null) System.out.println("Please also send the file " + logFile + " with your problem report.\n");
			System.out.println("Error localized message: " + e.getLocalizedMessage());
			System.out.println("Error toString: " + e.toString());
			System.out.println("Error getMessage: " + e.getMessage());
			e.printStackTrace();
			Statics.logException("main", e);
			System.out.println(Statics.startupLine);
			System.out.println(Version.getBuild());
			System.out.println("");
			Statics.outputVersionVendorUrl("");
			Statics.outputVersionVendorUrl("vm.specification.");
			Statics.outputVersionVendorUrl("vm.");
			Statics.outputVersionVendorUrl("specification.");
			System.out.println("Class version/path: " + System.getProperty("java.class.version") + " " + System.getProperty("java.class.path"));
			System.out.println("Library path: " + System.getProperty("java.library.path"));
			System.out.println("OS name/arch/version: " + System.getProperty("os.name") + " " + System.getProperty("os.arch") + " " + System.getProperty("os.version"));
			System.out.println("# available processors: " + Runtime.getRuntime().availableProcessors());
			System.out.println("Free memory in the VM: " + Runtime.getRuntime().freeMemory() + ", total memory in the VM: " + Runtime.getRuntime().maxMemory());

			System.exit(127);
		}
	}
}
