package com.vanheusden.pfa;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.*;

abstract class Drone implements Runnable {
	boolean canStop = false;
	BufferedReader input = null;
	BufferedWriter output = null;
	OutputStreamWriter outputSW = null;
	//
	Scene scene = null;
	PlayerColor color = null;
	int nr = -1;
	long deadLine = -1;
	//
	int latency = 0;	// in ms
	String name = null;
	final int sleepAfterQuit = 100; // ms
	final long datacommFailureSleepMin = 1, datacommFailureSleepMax = 512; // ms
	int nCores = 0;
	Map<String, String> parameters = null;
	//
	String host = null;
	int port = 55223;
	Socket s = null;
	boolean isSocket = false;
	//
	String exe = null, path = null;
	Process p = null;
	//
	Thread t = null;
	//
	Semaphore initScene = new Semaphore(1, false);
	//
	Semaphore moveLock = new Semaphore(1, false);
	Move move = null;
	AtomicBoolean initted = new AtomicBoolean(false);
	AtomicBoolean askTerminate = new AtomicBoolean(false);
	AtomicBoolean askedAnything = new AtomicBoolean(false);
	ArrayBlockingQueue<MoveResult> moveResultQueue = new ArrayBlockingQueue<MoveResult>(1);
	PMQuestioner pmq = null;
	AtomicInteger reconnects = new AtomicInteger(0);
	//
	AtomicLong lastNps = new AtomicLong(-1);
	long startAskTs = -1, askReplyTs = -1, askedDuration = -1;
	//
	Object notifyObject = null;
	//
	AtomicInteger resTimeOut = new AtomicInteger(0);
	AtomicInteger resChosen = new AtomicInteger(0);
	AtomicInteger resNotChosen = new AtomicInteger(0);
	AtomicInteger resIllegalMove = new AtomicInteger(0);
	AtomicInteger resNoIdea = new AtomicInteger(0);
	//
	AtomicBoolean pondering = new AtomicBoolean(false);
	//
	IO io;
	//
	int depth = -1, cp = -1, mate = -1, duration = -1, nps = -1;
	long nodes = -1;
	String pv = null;

	Drone(final String host, final int port, final boolean canStop, final Map<String, String> parameters) throws InterruptedException {
		io = new IO_logonly();
		logIgnoreIOException("New drone on " + host + ":" + port);

		this.host = host;
		this.port = port;
		this.canStop = canStop;
		isSocket = true;

		this.parameters = parameters;

		// this hardcoded value should be configurable or be measured
		// at run-time
		latency = 5;

		name = host + ":" + port;
	}

	Drone(final String exe, final String path, final boolean canStop, final Map<String, String> parameters) throws InterruptedException {
		io = new IO_logonly();
		logIgnoreIOException("New drone at " + exe);

		this.exe = exe;
		this.path = path;
		this.canStop = canStop;
		isSocket = false;

		this.parameters = parameters;

		latency = 0;

		name = exe;
	}

	void setIO(final IO io) {
		this.io = io;
	}

	abstract void stopPonder();
	abstract void startPonder(final Scene sc, final PlayerColor c, final Move lastMove);

	boolean getVerbose() {
		return Statics.getDebug() || isYes(parameters.get("debug"));
	}

	void countResult(PMResult r) {
		if (r == PMResult.timeOut)
			resTimeOut.addAndGet(1);
		else if (r == PMResult.chosen)
			resChosen.addAndGet(1);
		else if (r == PMResult.notChosen)
			resNotChosen.addAndGet(1);
		else if (r == PMResult.illegalMove)
			resIllegalMove.addAndGet(1);
		else if (r == PMResult.noIdea)
			resNoIdea.addAndGet(1);
	}

	int getEval() {
		return cp;
	}

	int getDepth() {
		return depth;
	}

	int getNTimeOut() {
		return resTimeOut.get();
	}

	int getNChosen() {
		return resChosen.get();
	}

	int getNNotChosen() {
		return resNotChosen.get();
	}

	int getNIllegalMove() {
		return resIllegalMove.get();
	}

	int getNoIdea() {
		return resNoIdea.get();
	}

	int getNReconnects() {
		return reconnects.get();
	}

	boolean isReady() {
		return initted.get();
	}

	boolean isThreadRunning() {
		if (t == null)
			return false;

		if (t.getState() == Thread.State.TERMINATED || t.getState() == Thread.State.NEW)
			return false;

		return true;
	}

	boolean isYes(String in) {
		if (in == null)
			return false;

		if (in.equalsIgnoreCase("yes") || in.equalsIgnoreCase("true") || in.equals("1"))
			return true;

		return false;
	}

	void startThread(final Object notifyObject) {
		this.notifyObject = notifyObject;

		assert t == null;
		t = new Thread(this);
		t.start();
	}

	void mySleepDuration(long duration) throws InterruptedException {
		logIgnoreIOException(getName() + " sleep for " + duration + "ms");

		Thread.sleep(duration);
	}

	void mySleepUntil(long until) throws InterruptedException {
		long duration = Statics.getDelayUntil(until);

		logIgnoreIOException(getName() + " sleep for " + duration + "ms");

		Thread.sleep(duration);
	}

	public void run() {
		for(;;) {
			if (askTerminate.get()) {
				logIgnoreIOException(getName() + " terminate thread: asked to terminate");
				break;
			}

			try {
				connect();

				if (input == null) {
					Thread.sleep(100);
					continue;
				}

				String line = input.readLine();

				if (line == null) {
					logIgnoreIOException(getName() + " null line");
					
					disconnect();
					input = null;
				}
				else {
					if (getVerbose())
						logIgnoreIOException(getName() + "> " + line);

					MoveResult temp = checkInput(line);

					if (temp != null) {
						if (moveResultQueue.offer(temp) == false)
							logIgnoreIOException(getName() + " return queue already had an item!");

						synchronized(notifyObject) {
							notifyObject.notify();
						}
					}
				}
			}
			catch(SocketException se) {
				logIgnoreIOException(getName() + " socket error " + se);
				disconnect();
				input = null;
			}
			catch(InterruptedException ie) {
				logIgnoreIOException(getName() + " interrupted " + ie);
			}
			catch(Exception e) {
				logIgnoreIOException(getName() + " (drone thread) exception " + e);
				Statics.logException("Drone", e);

				disconnect();
				input = null;
			}
		}

		logIgnoreIOException(getName() + " terminating thread");
	}

	abstract MoveResult checkInput(final String in);

	public MoveResult getResult(long before) throws IOException, InterruptedException {
		return moveResultQueue.poll(Statics.getDelayUntil(before), TimeUnit.MILLISECONDS);
	}

	public void destructor() throws IOException {
		logIgnoreIOException(getName() + " destructor, thread: " + (t != null));

		stopPonder();

		logIgnoreIOException(getName() + " MSTATS timeout: " + getNTimeOut() + ", chosen: " + getNChosen() + ", not chosen: " + getNNotChosen() + ", illegal moves: " + getNIllegalMove() + ", no idea: " + getNoIdea());

		askTerminate.set(true);

		if (t != null)
			t.interrupt();

		if (s != null)
			s.close();

		if (pmq != null)
			pmq.destructor();
	}

	public String getName() {
		if (name != null)
			return name;

		if (host != null)
			return "[" + host + "]:" + port;

		if (exe != null)
			return exe;

		Thread thread = Thread.currentThread();

		return thread.getName() + "[" + thread.getId() + "]";
	}

	public String toString() {
		return getName();
	}

	void setNCores(final int n) {
		nCores = n;
	}

	abstract void initSession() throws IOException, InterruptedException;

	protected void connect() throws InterruptedException {
		long curSleep = datacommFailureSleepMin;

		if (input == null)
			reconnects.addAndGet(1);

		while(input == null && !Thread.interrupted()) {
			initted.set(false);

			try {
				if (isSocket) {
					logIgnoreIOException(getName() + " Connecting to " + host + ":" + port);

					s = new Socket(host, port);
					s.setPerformancePreferences(0, 2, 1);
					s.setTcpNoDelay(true);

					logIgnoreIOException(getName() + " Connection established");

					if (s.getReceiveBufferSize() < 65536)
						s.setReceiveBufferSize(65536);

					input = new BufferedReader(new InputStreamReader(s.getInputStream()));

					outputSW = new OutputStreamWriter(s.getOutputStream());
					output = new BufferedWriter(outputSW);
				}
				else {
					logIgnoreIOException(getName() + " Starting " + exe);

					// FIXME set path
					p = Runtime.getRuntime().exec(exe);

					input = new BufferedReader(new InputStreamReader(p.getInputStream()));

					outputSW = new OutputStreamWriter(s.getOutputStream());
					output = new BufferedWriter(outputSW);
				}

				logIgnoreIOException(getName() + " Init session");

				initSession();

				if (input != null) {
					initted.set(true);

					synchronized(initted) {
						initted.notifyAll();
					}

					logIgnoreIOException(getName() + ": CONNECTED!");
				}
			}
			catch(ConnectException ce) {
				logIgnoreIOException(getName() + " failed to connect: " + ce);

				disconnect();
				input = null;

				mySleepDuration(curSleep);
			}
			catch(SocketException se) {
				logIgnoreIOException(getName() + " socket error " + se);
				disconnect();
				input = null;
			}
			catch(IOException ie) {
				logIgnoreIOException(getName() + " I/O exception " + ie);

				disconnect();
				input = null;

				mySleepDuration(curSleep);
			}

			if (input == null && curSleep < datacommFailureSleepMax)
				curSleep *= 2;
		}
	}

	void logIgnoreIOException(final String what) {
		try {
			io.progressCallbackDebug(what);
		}
		catch(IOException ie) {
			// ignore
		}
	}

	void disconnect() {
		if (pondering.getAndSet(false) || askedAnything.get()) {
			logIgnoreIOException(getName() + " disconnect");

			pmq.clearQueue();

			moveLock.acquireUninterruptibly();

			askedAnything.set(false);

			if (isSocket) {
				if (s != null) {
					try {
						s.close();
					}
					catch(IOException ie) {
						logIgnoreIOException(getName() + " Ignoring exception during disconnect " + ie);
					}
				}
			}
			else {
				if (p != null)
					p.destroy();
			}

			move = null;
			moveLock.release();

			if (t != null)
				t.interrupt();
		}
		else {
			logIgnoreIOException(getName() + " ignoring disconnect because nothing was asked");
		}
	}

	void emit(String str) throws IOException {
		if (output != null) {
			output.write(str + "\n");

			output.flush();

			if (outputSW != null)
				outputSW.flush();

			logIgnoreIOException(getName() + "> " + str);
		}
		else {
			logIgnoreIOException(getName() + "/NULL> " + str);
		}
	}

	class PMQuestioner implements Runnable {
		class PMQuestion {
			final Move lastMove;

			PMQuestion(final Move lastMove) {
				this.lastMove = lastMove;
			}

			Move getLastMove() {
				return lastMove;
			}
		}

		ArrayBlockingQueue<PMQuestion> moveResult = new ArrayBlockingQueue<PMQuestion>(1);
		Thread t = null;

		PMQuestioner() {
			t = new Thread(this);
			t.start();
		}

		void destructor() {
			if (t != null)
				t.interrupt();
		}

		void clearQueue() {
			moveResult.clear();
		}

		void startAsk(final Move lastMove) {
			logIgnoreIOException(getName() + " scheduling question");

			moveResult.offer(new PMQuestion(lastMove));
		}

		void doAsk(PMQuestion q) {
			logIgnoreIOException("PMQuestion must be overridden");
		}

		public void run() {
			for(;!Thread.interrupted();) {
				try {
					PMQuestion q = moveResult.poll(3600, TimeUnit.SECONDS);
					if (q == null)
						continue;

					logIgnoreIOException(getName() + " got a question");

					if (initted.get())
						doAsk(q);
					else {
						logIgnoreIOException(getName() + " Drone not initialized");

						initScene.acquireUninterruptibly();
						long useDeadLine = deadLine - latency;
						initScene.release();

						long timeLeft = Statics.getDelayUntil(useDeadLine);
						long maxWaitTillConnected = timeLeft / 4; // arbitrary guess estimate

						if (maxWaitTillConnected > 0) {
							logIgnoreIOException(getName() + " waiting " + maxWaitTillConnected + "ms");

							synchronized(initted) {
								initted.wait(maxWaitTillConnected);
							}

							if (initted.get()) {
								logIgnoreIOException(getName() + " Drone initialized, asking question");
								doAsk(q);
							}
							else {
								logIgnoreIOException(getName() + " Drone still not initialized");
							}
						}
					}
				}
				catch(InterruptedException ie) {
					logIgnoreIOException("PMQuestion thread interrupted");
					break;
				}
			}
		}
	}

	public abstract void startGame(final int nMoves, final long maxDuration) throws IOException, InterruptedException;

	public void startAsk(final Scene sc, final PlayerColor c, final Move lastMove, final int nr, final long deadLine) {
		startAskTs = System.currentTimeMillis();

		initScene.acquireUninterruptibly();
		scene = sc;
		color = c;
		this.nr = nr;
		this.deadLine = deadLine;
		initScene.release();

		askedDuration = Statics.getDelayUntil(deadLine);

		pmq.startAsk(lastMove);
	}

	public abstract void stopAsk();

	public abstract void interruptThinking();

	long peekLastNps() {
		return lastNps.get();
	}

	long getLastNps() {
		long temp = lastNps.get();
		if (System.currentTimeMillis() >= deadLine)
			lastNps.set(-1);
		return temp;
	}

	long peekThinkDuration() {
		if (startAskTs > 0)
			return askReplyTs - startAskTs;

		return -1;
	}

	long getThinkDuration() {
		if (startAskTs > 0) {
			long took = askReplyTs - startAskTs;

			askedDuration = startAskTs = askReplyTs = -1;

			return took;
		}

		return -1;
	}
}
