package com.vanheusden.pfa;

import java.io.*;
import java.util.*;

class IO_Tournament extends IO {
	final String boardInit;
	List<Drone> drones;
	int maxNMoves = -1;

	public IO_Tournament(final String pgnFile, final int maxDepth, final int maxExtension, final boolean random, final double maxDuration, final String boardInit, final List<Drone> drones, final int maxNMoves, final boolean ponder, List<String> beepers) throws Exception {
		super(maxDepth, maxExtension, random, maxDuration, null, ponder, false, beepers);

		this.boardInit = boardInit;

		this.pgnFile = pgnFile;

		this.drones = drones;

		if (maxNMoves == -1) {
			this.maxNMoves = 40;
			Statics.log("Setting default maximum number of moves per game to " + this.maxNMoves);
		}
		else {
			this.maxNMoves = maxNMoves;
		}

		for(Drone current : drones)
			current.startThread(new Object());
	}

	void output(String str) {
		Statics.log(str);

		System.out.println(Statics.getTimestampStr() + " " + str);
	}

	public void loop() throws IOException, InterruptedException {
		int nDrones = drones.size();

		// randomize order so that short runs don't always touch the same engines
		if (nDrones >= 3) {
			for(int index=0; index<nDrones; index++) {
				int index2 = -1;

				do {
					index2 = r.nextInt(nDrones);
				}
				while(index == index2);

				Drone dummy = drones.get(index);
				drones.set(index, drones.get(index2));
				drones.set(index2, dummy);
			}
		}

		for(int p1=0; p1<nDrones; p1++) {
			for(int p2=0; p2<nDrones; p2++) {
				Drone pWhite = drones.get(p1);
				Drone pBlack = drones.get(p2);

				Statics.log("Matching " + pWhite.getName() + " against " + pBlack.getName());

				if (p1 == p2) {
					Statics.log("Skipping...");
					continue;
				}

				Scene scene = null;
				if (boardInit != null) {
					Object [] epd = Statics.EPDToScene(boardInit);
					scene = (Scene)epd[0];
					currentPlayer = (PlayerColor)epd[1];
				}
				else {
					scene = new Scene();
					currentPlayer = PlayerColor.WHITE;
				}

				output(pWhite.getName() + " versus " + pBlack.getName());

				Pgn pgnObject = new Pgn(pWhite.getName(), pBlack.getName(), pgnFile);

				long timeLeftWhite = (long)(maxDuration * 1000.0);
				long timeLeftBlack = (long)(maxDuration * 1000.0);

				pWhite.startGame(maxNMoves, timeLeftWhite);
				pBlack.startGame(maxNMoves, timeLeftBlack);

				PlayerColor winner = null;
				MoveResult lastMove = null;
				int nr = 1;
				boolean crash = false;
				for(;;) {
					MoveResult move = null;
					String curName;

					long startedOn = System.currentTimeMillis(), took = -1;

					int movesLeft = 0;
					if (maxNMoves > 0)
						movesLeft = maxNMoves - nr + 1;

					if (currentPlayer == PlayerColor.WHITE) {
						curName = pWhite.getName();

						try {
							pWhite.startAsk(scene, currentPlayer, lastMove.getMove(), movesLeft, startedOn + timeLeftWhite);
							move = pWhite.getResult(startedOn + timeLeftWhite);

							took = System.currentTimeMillis() - startedOn;
							timeLeftWhite -= took;
						}
						catch(IOException ie) {
							crash = true;
							output("result: engine " + curName + " crashed");
							break;
						}
					}
					else {
						curName = pBlack.getName();

						try {
							pBlack.startAsk(scene, currentPlayer, lastMove.getMove(), movesLeft, startedOn + timeLeftBlack);
							move = pBlack.getResult(startedOn + timeLeftBlack);

							took = System.currentTimeMillis() - startedOn;
							timeLeftBlack -= took;
						}
						catch(IOException ie) {
							crash = true;
							output("result: engine " + curName + " crashed");
							break;
						}
					}

					if (move == null) {
						output("result: engine " + curName + " did not return a move (in time)");
						winner = Statics.opponentColor(currentPlayer);
						break;
					}

					lastMove = move;

					if (currentPlayer == PlayerColor.WHITE)
						output("" + nr + ". " + move + " (" + took + "ms)");
					else
						output("" + nr + ". ... " + move + " (" + took + "ms)");

					pgnObject.pgnEmitMove(scene.getBoard(), move.getMove(), currentPlayer);

					scene = scene.Move(move.getMove());

					currentPlayer = Statics.opponentColor(currentPlayer);

					scene.validateMoves(currentPlayer);

					if (timeLeftWhite <= 0) {
						output("White has no time left (" + timeLeftWhite + ")");
						winner = PlayerColor.BLACK;
						break;
					}

					if (timeLeftBlack <= 0) {
						output("Black has no time left (" + timeLeftBlack + ")");
						winner = PlayerColor.WHITE;
						break;
					}

					if (scene.isCheckMate(currentPlayer)) {
						output("result: checkmate");
						winner = Statics.opponentColor(currentPlayer);
						break;
					}

					if (scene.isStaleMate(PlayerColor.WHITE) || scene.isStaleMate(PlayerColor.BLACK)) {
						output("result: stalemate");
						break;
					}

					if (scene.isInsufficientMaterialDraw()) {
						output("result: insufficient material draw");
						break;
					}

					if (currentPlayer == PlayerColor.WHITE)
						nr++;

					if (maxNMoves > 0 && nr == maxNMoves) {
						output("result: maximum number of moves (" + maxNMoves + ") reached");
						break;
					}
				}

				if (!crash) {
					String result = null;

					if (winner == null)
						result = "1/2-1/2";
					if (winner == PlayerColor.WHITE)
						result = "1-0";
					else if (winner == PlayerColor.BLACK)
						result = "0-1";

					output("result: " + result);

					pgnObject.closePgnFile(result);
				}
			}
		}

		for(Drone current : drones)
			current.destructor();
	}
}
