package com.vanheusden.pfa;

import java.io.*;
import java.util.*;

class IO_UCI extends IO {
	BufferedReader xboardPipe = null;
	BufferedWriter output = null;
	String pgnFile;
	int it = 0;
	Move lastMove = null;
	boolean uciAllowPonder = false;
	//
	final int defaultNMovesPerGame = 40;
	boolean isAsIsThinkTime = true;
	long thinkTime = 5000;
	int maxNMoves = defaultNMovesPerGame, timeInc = 0;

	public void progressCallbackDebug(String what) throws IOException {
		if (Statics.verbose > 0) {
			Statics.log(what);
			emit("info string " + what);
		}
	}

	public void progressCallback(String what) throws IOException {
		Statics.log(what);

		if (Statics.verbose > 0)
			emit("info string " + what);
	}

	public void progressCallbackNPS(int nps) throws IOException {
		if (Statics.verbose > 0)
			emit("info nps " + nps);
	}

	public void progressCallbackTimeSearch(int ms) throws IOException {
		if (Statics.verbose > 0)
			emit("info time " + ms);
	}

	public void progressCallbackBest(Move move, int score) throws IOException {
		if (Statics.verbose > 0 && move != null) {
			emit("info pv " + move);
			emit("info score cp " + score);
		}
	}

	public void progressCallbackNNodes(long nNodes) throws IOException {
		if (Statics.verbose > 0)
			emit("info nodes " + nNodes);
	}

	public void progressCallbackDepthReached(int depth) throws IOException {
		if (Statics.verbose > 0)
			emit("info depth " + depth);
	}

	public void progressCallbackCurrMove(Move move) throws IOException {
		if (Statics.verbose > 0)
			emit("info currmove " + move);
	}

	public void progressCallbackXboardShowThinking(int depth, int score, long durationCentiSeconds, long nodesVisited, String pv) throws IOException {
		progressCallbackTimeSearch((int)(durationCentiSeconds * 10));
		progressCallbackDepthReached(depth);

		if (Statics.verbose > 0) {
			emit("info score cp " + score);
			emit("info pv " + pv);
		}
	}

        private void restartGame() {
                scene = new Scene();
                moveRepHist = new MoveRepetition();
                currentPlayer = PlayerColor.WHITE;
                it = 1;
		lastMove = null;
        }

	void switchUser() {
		currentPlayer = Statics.opponentColor(currentPlayer);
		if (currentPlayer == PlayerColor.WHITE) {
			it++;
		}
	}

	void setupIO() throws IOException {
		xboardPipe = new BufferedReader(new InputStreamReader(System.in));
		output = new BufferedWriter(new OutputStreamWriter(System.out));
	}

	synchronized void emit(String what) throws IOException {
		Statics.log("Output: <" + what + ">");

		output.write(what);
		output.newLine();
		output.flush();
	}

	String get() throws IOException {
		return xboardPipe.readLine();
	}

	IO_UCI(String pgnFile, int maxDepth, int maxExtension, boolean random, double maxDuration, final Brain brain, final boolean ponder, final boolean ignore_protocol_overrides, List<String> beepers) throws Exception {
		super(maxDepth, maxExtension, random, maxDuration, brain, ponder, ignore_protocol_overrides, beepers);

		setupIO();

		this.pgnFile = pgnFile;

		restartGame();
	}

	public void UCI_err(String error) throws IOException {
		Statics.log(error);
		emit("info string " + error);
	}

	void doEngineMove() throws IOException, InterruptedException {
		progressCallback("Start search");

		double currentDuration = thinkTime / 1000.0;

		if (isAsIsThinkTime == false) {
			int movesLeft = maxNMoves - it;
			int curNMoves = movesLeft <= 0 ? defaultNMovesPerGame : movesLeft;
			currentDuration = 0.6 * (thinkTime / 1000.0 + (curNMoves - 1) * timeInc) / (curNMoves + 7);

			double limitDuration = thinkTime / (15.0 * 1000.0);
			if (currentDuration > limitDuration)
				currentDuration = limitDuration;
		}

		if (currentDuration > maxDuration && maxDuration > 0) {
			Statics.log("Selected think-time longer than max (" + currentDuration + "), limiting to " + maxDuration);
			currentDuration = maxDuration;
		}

		if (bp != null)
			bp.stopPonder();

		String moveStr = null;
		MoveResult move = brain.calculateMove(it, scene, maxDepth, currentPlayer, moveRepHist, (long)(currentDuration * 1000.0), 0, lastMove, false);

		scene = scene.Move(move.getMove());

		switchUser();

		scene.validateMoves(currentPlayer);

		if (bp != null && ponder && (uciAllowPonder || ignore_protocol_overrides))
			bp.startPonder(it, scene, maxDepth, currentPlayer, moveRepHist, 0, 0, move.getMove());

		moveStr = Statics.moveToString(move.getMove(), true);

		emit("info score cp " + move.getEval());
		emit("bestmove " + moveStr);
	}

	public void loop() throws IOException, InterruptedException {
		boolean searching = false;

		Statics.log(Statics.startupLine);
		Statics.log("Pos in uci mode started");
		
		int nProcs = Runtime.getRuntime().availableProcessors();
		progressCallback("# cpus: " + nProcs);

		brain.setName(System.getProperty("user.name") + "@localhost");

		int movesDone = 0;
		for(;;) {
			String cmd, par = null;
			String [] values;
			String line = get();
			if (line == null) {
				UCI_err("Got null!");
				break;
			}

			Statics.log("In: <" + line + ">");
			line = line.trim();
			line = line.replaceAll(" +", " ");

			if (line.length() == 0) {
				UCI_err("ignoring empty line");
				continue;
			}

			values = line.split(" ");

			cmd = values[0];
			if (values.length >= 2) {
				par = values[1];
			}

			if (cmd.equalsIgnoreCase("quit")) {
				Statics.log("quiting");
				break;
			}
			else if (cmd.equalsIgnoreCase("UCI")) {
				emit("id name " + Version.getIdentifier());
				emit("id author Folkert van Heusden <folkert@vanheusden.com>");
				emit("option name UCI_Opponent type string default <empty>");
				emit("option name UCI_EngineAbout type string default http://www.vanheusden.com/DeepBrutePos/");
				emit("uciok");
			}
			else if (cmd.equalsIgnoreCase("debug")) {
				// yeah well
				UCI_err("ignoring debug command");
			}
			else if (cmd.equalsIgnoreCase("isready")) {
				emit("readyok");
			}
			else if (cmd.equalsIgnoreCase("ucinewgame")) {
				restartGame();
			}
			else if (cmd.equalsIgnoreCase("position")) {
				int index = -1;

				restartGame();	// implicit(!)

				if (values[1].equalsIgnoreCase("fen")) {
					Object [] epd = Statics.EPDToScene(values[2] + " " + values[3] + " " + values[4] + " " + values[5]);
					scene = (Scene)epd[0];
					currentPlayer = (PlayerColor)epd[1];

					index = 9;
				}
				else if (values[1].equalsIgnoreCase("startpos")) {
					// values[2] == 'moves'

					index = 3;
				}
				else {
					UCI_err("invalid position line");
				}
				scene.validateMoves(currentPlayer);

				// do moves from values[index] ... values[values.length - 1]
				int count = 0;
				for(;index<values.length; index++) {
					String moveString = values[index];
					StringBuffer error = new StringBuffer("");
					Move move = Statics.LANToMove(moveString, scene.getBoard(), currentPlayer, scene.getMoveList(currentPlayer), error);
					if (move == null) {
						String errorMessage = "Illegal move (" + error + "): " + moveString;
						UCI_err(errorMessage);
					}
					else {
						moveRepHist.push(currentPlayer, brain.getHash(scene.getBoard(), currentPlayer));

						scene = scene.Move(move);
						switchUser();
						scene.validateMoves(currentPlayer);
						count++;
					}

					lastMove = move;
				}

				movesDone = count / 2;
			}
			else if (cmd.equalsIgnoreCase("setoption")) {
				if (!values[1].equalsIgnoreCase("name")) {
					UCI_err("expecting 'name' behind 'setoption'");
					continue;
				}

				if (values[2].equalsIgnoreCase("UCI_Opponent")) {
					String opponent = "";
					for(int index=4; index<Math.min(8, values.length); index++)
						opponent += values[index] + " ";
					Statics.log("playing against: " + opponent);
					brain.setName(opponent);

				}
				else if (values[2].equalsIgnoreCase("random")) {
					if (values[4].equalsIgnoreCase("true")) {
						Statics.log("enabling randomizer");
						brain.setAllowRandom(true);
					}
					else {
						Statics.log("disabling randomizer");
						brain.setAllowRandom(false);
					}
				}
				else {
					Statics.log("unknown option: " + line);
				}
			}
			else if (cmd.equalsIgnoreCase("register")) {
				Statics.log("ignoring register-line: " + line);
			}
			else if (cmd.equalsIgnoreCase("go")) {
				//int nToGo = Math.max(0, 50 - movesDone);
				// double time = maxDuration * nToGo;
				double time = maxDuration;
				for(int index=1; index<values.length; index++) {
					// go wtime 300000 btime 300000 movestogo 40
					if ((values[index].equalsIgnoreCase("wtime") && currentPlayer == PlayerColor.WHITE) ||
					    (values[index].equalsIgnoreCase("btime") && currentPlayer == PlayerColor.BLACK)) {
						thinkTime = Long.valueOf(values[++index]);
						isAsIsThinkTime = false;
					}
					else if (values[index].equalsIgnoreCase("movestogo")) {
						maxNMoves = Integer.valueOf(values[++index]);
						isAsIsThinkTime = false;
					}
					else if (values[index].equalsIgnoreCase("movetime")) {
						thinkTime = Long.valueOf(values[++index]);
						isAsIsThinkTime = true;
					}
					else if (values[index].equalsIgnoreCase("ponder")) {
						uciAllowPonder = true;
					}
				}

				doEngineMove();
			}
			else if (cmd.equalsIgnoreCase("stop")) {
				if (!searching) {
					UCI_err("Were not searching when receiving 'stop' command?!");
				}
				else {
					Statics.log("Was searching for a while");
					searching = false;
					doEngineMove();
				}
			}
			else {
				UCI_err("command not understood: " + line);
			}
		}
	}
}
