package com.vanheusden.pfa;

import java.util.*;
import java.io.*;

public class Statics {
	static String startupLine = null;
	static String logFile;
	static int verbose = 0;
	static boolean debug = false;

	static void updateStartupLine() {
		startupLine = "*** " + Version.getIdentifier() + " (C) 2006-2015 by folkert@vanheusden.com ***";
	}

	static void setDebug(final boolean debugIn) {
		debug = debugIn;
	}

	static boolean getDebug() {
		return debug;
	}

	static long getDelayUntil(long until) {
		return Math.max(0, until - System.currentTimeMillis());
	}

	static String getTimestampStr() {
		long ts = System.currentTimeMillis();

		Calendar c = Calendar.getInstance();
		c.setTimeInMillis(ts);

		return String.format("%04d-%02d-%02d %02d:%02d:%02d.%03d",
			c.get(Calendar.YEAR), c.get(Calendar.MONTH) + 1, c.get(Calendar.DAY_OF_MONTH),
			c.get(Calendar.HOUR_OF_DAY), c.get(Calendar.MINUTE), c.get(Calendar.SECOND),
			ts % 1000);
	}

	static void log(final String what) {
		if (logFile != null || debug) {
			String out = getTimestampStr() + " " + what;

			if (logFile != null) {
				try {
					FileWriter logFileHandle = new FileWriter(logFile, true);
					PrintWriter logFileHandle2 = new PrintWriter(logFileHandle);

					logFileHandle2.println(out);

					logFileHandle2.close();
					logFileHandle.close();
				}
				catch(Exception e) {
					System.err.println("Failed to write to logfile " + logFile);
				}
			}

			if (debug)
				System.err.println(out);
		}
	}

	public static void logException(final String note, final Exception e) {
		if (note != null)
			log(note + " caught exception: " + e);
		else
			log("Caught exception: " + e);

		log("Details: " + e.getMessage());
		log("Stack-trace:");

		for(StackTraceElement ste: e.getStackTrace()) {
			log(" " + ste.getClassName() + ", " + ste.getFileName() + ", " + ste.getLineNumber() + ", " + ste.getMethodName() + ", " + (ste.isNativeMethod() ? "is native method" : "NOT a native method"));
		}
//System.exit(127); // FIXME is temp
	}

	public static String [] genBoard(final Board board) {
		final String [] asciiBoard = new String[8 * 2 + 2];
		int index = 0;

		for(int y=7; y>=0; y--) {
			asciiBoard[index++] = "  ---+---+---+---+---+---+---+---";

			String line = "" + (char)('1' + y) + " ";
			for(int x=0; x<8; x++) {
				ChessObject currentField = board.getAt(x, y);

				if (currentField != null) {
					ChessObjectType type = currentField.getType();
					char color = ' ';

					if (currentField.getColor() == PlayerColor.BLACK)
						color = '*';

					line += ""+color;

					if (type == null)
						line += " ";
					else
						line += Statics.objectTypeToString(type, true);

					line += ""+color;
				}
				else {
					line += "   ";
				}

				if (x != 7)
					line += "|";
			}

			asciiBoard[index++] = line;
		}
		asciiBoard[index++] = "  ---+---+---+---+---+---+---+---";
		asciiBoard[index++] = "   A | B | C | D | E | F | G | H ";

		return asciiBoard;
	}

	public static void displayBoard(final Board board) {
		String [] asciiBoard = genBoard(board);

		for(String line : asciiBoard) {
			System.out.println(line);
		}
	}

	public static void logBoard(final Board board) {
		String [] asciiBoard = genBoard(board);

		for(String line : asciiBoard) {
			Statics.log(line);
		}
	}

	public static void setLogFile(final String logFile) {
		Statics.logFile = logFile;
	}

	public static void increaseVerbosity() {
		verbose++;
	}

	public static int getVerbosity() {
		return verbose;
	}

	public static PlayerColor opponentColor(final PlayerColor color) {
		if (color == PlayerColor.WHITE)
			return PlayerColor.BLACK;

		return PlayerColor.WHITE;
	}

	// FIXME: redundant
	public static String objectTypeToString(final ChessObjectType type, final boolean uppercase) {
		String result = null;

		switch(type) {
			case QUEEN:
				result = "q";
				break;

			case KNIGHT:
				result = "n";
				break;

			case BISHOP:
				result = "b";
				break;

			case ROOK:
				result = "r";
				break;

			case PAWN:
				result = "p";
				break;

			case KING:
				result = "k";
				break;

			default:
				assert false;
		}

		if (uppercase)
			result = result.toUpperCase();

		return result;
	}

	public static String moveToString(final Move move, final boolean xboardMode) {
		if (move.getMoveType() == MoveType.CASTLING) {
			if (move.getTX() == 6) {
				if (xboardMode)
					return "" + (move.getFY() == 0?"e1g1":"e8g8");
				else
					return "O-O";
			}
			else {
				if (xboardMode)
					return "" + (move.getFY() == 0?"e1c1":"e8c8");
				else
					return "O-O-O";
			}
		}
		else if (move.getMoveType() == MoveType.EN_PASSANT) {
			return ""+(char)('a' + move.getFX())+(char)('1' + move.getFY()) + (char)('a' + move.getTX())+(char)('1' + move.getTY());
		}
		else {
			String moveStr = ""+(char)('a' + move.getFX())+(char)('1' + move.getFY()) + (xboardMode == false?" to ":"") + (char)('a' + move.getTX())+(char)('1' + move.getTY());

			if (move.getPromoteTo() != null)
				moveStr += objectTypeToString(move.getPromoteTo(), false);

			return moveStr;
		}
	}

	public static void outputVersionVendorUrl(final String prefix) {
		System.out.println("Java " + prefix + " version/vendor/url: " + 
				System.getProperty("java." + prefix + "version") + " " + 
				System.getProperty("java." + prefix + "vendor") + " " + 
				System.getProperty("java." + prefix + "vendor.url")
				);
	}

	public static double getTimestamp() {
		return System.currentTimeMillis() / 1000.0;
	}

	public static double timeLeft(final double finishAt) {
		double left = finishAt - getTimestamp();

		if (left >= 0.0)
			return left;

		return 0.0;
	}

	public static Object [] EPDToScene(final String fenPos) {
		String [] values = fenPos.split(" ");

		Board currentBoard = FENStringToBoard(values[0], values[1], values[2]);

		List<ChessObject> whiteTemp = new ArrayList<ChessObject>();
		List<ChessObject> blackTemp = new ArrayList<ChessObject>();

		boardToChessObjectArrays(currentBoard, whiteTemp, blackTemp);

		Move causingMove = null;

		if (values[3].equals("-") == false) {
			int dx = values[3].toLowerCase().charAt(0) - 'a';
			int dy = values[3].toLowerCase().charAt(1) - '1';
			int fx = dx;
			int fy = dy == 2 ? 1 : 6;
			int tx = dx;
			int ty = dy == 2 ? 3 : 4;

			causingMove = new Move(fx, fy, tx, ty);

			// no need to update the board as it won't be used anymore in
			// this method
			ChessObject o = currentBoard.getAt(tx, ty);
			o.doMove(fx, fy);
			o.resetNMoves(); // it is the first move
		}

		assert whiteTemp.size() > 0;
		assert blackTemp.size() > 0;

		ChessObject [] white = whiteTemp.toArray(new ChessObject[whiteTemp.size()]);

		ChessObject [] black = blackTemp.toArray(new ChessObject[blackTemp.size()]);

		PlayerColor player = values[1].equalsIgnoreCase("w") ? PlayerColor.WHITE : PlayerColor.BLACK;

		if (causingMove != null)
			return new Object [] { new Scene(white, black, causingMove), player };

		return new Object [] { new Scene(white, black), player };
	}

	public static Board FENStringToBoard(String piecePlacement, String activeColor, String castlingAvailability) {
		Board board = new Board();
		int y = 7;
		int x = 0;

		for(int index=0; index<piecePlacement.length(); index++) {
			char currentChar = piecePlacement.charAt(index);
			ChessObjectType objectType = null;
			PlayerColor objectColor = null;
			int currentX = x;
			int currentY = y;

			if (Character.isUpperCase(currentChar)) {
				objectColor = PlayerColor.WHITE;
			}
			else {
				objectColor = PlayerColor.BLACK;
			}

			currentChar = Character.toUpperCase(currentChar);
			if (currentChar == 'P') {
				objectType = ChessObjectType.PAWN;
				x++;
			}
			else if (currentChar == 'N') {
				objectType = ChessObjectType.KNIGHT;
				x++;
			}
			else if (currentChar == 'B') {
				objectType = ChessObjectType.BISHOP;
				x++;
			}
			else if (currentChar == 'R') {
				objectType = ChessObjectType.ROOK;
				x++;
			}
			else if (currentChar == 'Q') {
				objectType = ChessObjectType.QUEEN;
				x++;
			}
			else if (currentChar == 'K') {
				objectType = ChessObjectType.KING;
				x++;
			}
			else if (Character.isDigit(currentChar)) {
				x += Character.digit(currentChar, 10);
			}
			else if (currentChar == '/') {
				x = 0;
				y--;
			}

			if (objectType != null) {
				ChessObject currentObject;

				currentObject = ChessObject.newChessObject(objectType, currentX, currentY, objectColor);

				currentObject.setNMoves();

				if (objectColor == PlayerColor.WHITE && y == 1 && objectType == ChessObjectType.PAWN)
					currentObject.resetNMoves();
				else if (objectColor == PlayerColor.BLACK && y == 6 && objectType == ChessObjectType.PAWN)
					currentObject.resetNMoves();

				board.set(currentObject);
			}
		}

		for(int index=0; index<castlingAvailability.length(); index++) {
			char currentChar = castlingAvailability.charAt(index);

			if (currentChar == '-')
				break;

			if (currentChar == 'K') { // white king side => right
				y = 0;
				x = 7;
			}
			else if (currentChar == 'Q') { // white queen side => left
				y = 0;
				x = 0;
			}
			else if (currentChar == 'k') { // black king side => right
				y = 7;
				x = 7;
			}
			else if (currentChar == 'q') { // black queen side => left
				y = 7;
				x = 0;
			}
			else {
				assert 0 == 1;
			}

			board.getAt(x, y).resetNMoves();
			// if one can still castle with a rook, the king should also be able to castle
			board.getAt(4, y).resetNMoves();
		}

		return board;
	}

	public static void boardToChessObjectArrays(Board board, List<ChessObject> white, List<ChessObject> black) {
		int x, y;

		for(y=0; y<8; y++) {
			for(x=0; x<8; x++) {
				ChessObject currentObject = board.getAt(x, y);

				if (currentObject != null) {
					if (currentObject.getColor() == PlayerColor.WHITE) {
						white.add(currentObject);
					}
					else if (currentObject.getColor() == PlayerColor.BLACK) {
						black.add(currentObject);
					}
				}
			}
		}
	}

	public static boolean canRookStillCastle(Board board, int x, int y) {
		ChessObject rook = board.getAt(x, y);
		ChessObject king = board.getAt(4, y);

		if (king == null)
			return false;

		if (rook == null)
			return false;

		if (king.isFirstMove() == false)
			return false;

		if (rook.isFirstMove() == false)
			return false;

		if (y == 0 && rook.getColor() != PlayerColor.WHITE)
			return false;

		if (y == 0 && king.getColor() != PlayerColor.WHITE)
			return false;

		if (y == 7 && rook.getColor() != PlayerColor.BLACK)
			return false;

		if (y == 7 && king.getColor() != PlayerColor.BLACK)
			return false;

		return true;
	}

	public static String toFENString(Board board, PlayerColor playerOnMove, Move lastMove, int moveNr, final boolean skipClock) {
		ChessObject lastMoveObject = lastMove != null ? board.getAt(lastMove.getTX(), lastMove.getTY()) : null;
		String fs = "";
		boolean canCastle = false;

		for(int y=7; y>=0; y--) {
			int empty = 0;

			for(int x=0; x<8; x++) {
				if (board.getAt(x, y) == null) {
					empty++;
				}
				else {
					if (empty > 0) {
						fs += "" + empty;
						empty = 0;
					}

					fs += board.getAt(x, y).getTypeCharacter(board.getAt(x, y).getColor() == PlayerColor.WHITE);
				}
			}

			if (empty > 0) {
				fs += "" + empty;
			}

			if (y > 0) {
				fs += "/";
			}
		}

		if (playerOnMove == PlayerColor.WHITE)
			fs += " w";
		else
			fs += " b";

		fs += " ";
		if (canRookStillCastle(board, 7, 0) == true) {
			fs += "K";
			canCastle = true;
		}
		if (canRookStillCastle(board, 0, 0) == true) {
			fs += "Q";
			canCastle = true;
		}
		if (canRookStillCastle(board, 7, 7) == true) {
			fs += "k";
			canCastle = true;
		}
		if (canRookStillCastle(board, 0, 7) == true) {
			fs += "q";
			canCastle = true;
		}
		if (canCastle == false) {
			fs += "-";
		}

		if (lastMoveObject != null && lastMoveObject instanceof ChessObjectPawn && Math.abs(lastMove.getFY() - lastMove.getTY()) == 2) {
			fs += " " + (char)('a' + lastMove.getTX()) + (lastMoveObject.getColor() == PlayerColor.WHITE ? 3 : 6);
		}
		else {
			fs += " -";
		}

		//		fs += " " + (moveNr - 1);

		//		fs += " " + (moveNr * 2 + (playerOnMove == PlayerColor.BLACK ? 1 : 0));
		//

		if (!skipClock)
			fs += " 10 " + moveNr;

		return fs;
	}

	public static ChessObjectType CharToChessObjectType(char objectChar) {
		switch(objectChar) {
			case 'P':
			case 'p':
				return ChessObjectType.PAWN;

			case 'B':
			case 'b':
				return ChessObjectType.BISHOP;

			case 'R':
			case 'r':
				return ChessObjectType.ROOK;

			case 'N':
			case 'n':
				return ChessObjectType.KNIGHT;

			case 'Q':
			case 'q':
				return ChessObjectType.QUEEN;

			case 'K':
			case 'k':
				return ChessObjectType.KING;
		}

		return null;
	}

	public static boolean isUpper(char checkChar) {
		if (checkChar >= 'A' && checkChar <= 'Z') {
			return true;
		}

		return false;
	}

	public static boolean isDigit(char checkChar) {
		if (checkChar >= '0' && checkChar <= '9') {
			return true;
		}

		return false;
	}

	public static Move LANToMove(String moveString, Board board, PlayerColor color, List<Move> validMoves, StringBuffer error) {
		int baseLine = color == PlayerColor.WHITE ? 0 : 7;
		int fx, fy, tx, ty;
		MoveType moveType = MoveType.REGULAR;
		int maxStringLen;
		ChessObjectType promoteTo = null;
		ChessObject objectAtFrom = null;
		Move move = null;

		if (moveString.length() > 1 && isUpper(moveString.charAt(0)))
			moveString = moveString.substring(1);

		moveString = moveString.toLowerCase();

		if (moveString.compareToIgnoreCase("oo") == 0 || moveString.compareTo("o-o") == 0) {
			fy = ty = baseLine;
			fx = 4;
			tx = 6;
			moveType = MoveType.CASTLING;
			move = new Move(fx, fy, tx, ty, moveType);
		}
		else if (moveString.compareToIgnoreCase("ooo") == 0 || moveString.compareTo("o-o-o") == 0) {
			fy = ty = baseLine;
			fx = 4;
			tx = 2;
			moveType = MoveType.CASTLING;
			move = new Move(fx, fy, tx, ty, moveType);
		}
		else if (moveString.length() >= 4) {
			fx = moveString.charAt(0) - 'a';
			fy = moveString.charAt(1) - '1';
			if (moveString.charAt(2) == '-') {
				moveString = moveString.substring(0, 2) + moveString.substring(3);
			}
			tx = moveString.charAt(2) - 'a';
			ty = moveString.charAt(3) - '1';
			if (moveString.length() == 6 && moveString.charAt(4) == '=') {
				promoteTo = CharToChessObjectType(moveString.charAt(5));
				maxStringLen = 6;
			}
			else if (moveString.length() == 5) {
				promoteTo = CharToChessObjectType(moveString.charAt(4));
				maxStringLen = 5;
			}
			else {
				maxStringLen = 4;
			}

			if (fx < 0 || fy < 0 || tx < 0 || ty < 0 || fx > 7 || fy > 7 || tx > 7 || ty > 7) {
				error.append("Move " + moveString + ": invalid from/to characters");
				return null;
			}
			else if (moveString.length() != maxStringLen) {
				error.append("Move " + moveString + ": unexpected movestring length");
				return null;
			}

			objectAtFrom = board.getAt(fx, fy);
			if (objectAtFrom != null) {
				if (objectAtFrom instanceof ChessObjectKing && fy == baseLine && ty == baseLine && fx == 4) {
					if (tx == 6) { /* short castling */
						moveType = MoveType.CASTLING;
					}
					else if (tx == 2) { /* long castling */
						moveType = MoveType.CASTLING;
					}
				}
				else if (objectAtFrom instanceof ChessObjectPawn && fx != tx && board.getAt(tx, ty) == null) { // en passant?
					moveType = MoveType.EN_PASSANT;
				}
				else if (objectAtFrom instanceof ChessObjectPawn &&
						((color == PlayerColor.WHITE && ty == 7) ||
						 (color == PlayerColor.BLACK && ty == 0)) && promoteTo == null) {
					System.out.println("	  promotion without given to-type, assuming queen");
					promoteTo = ChessObjectType.QUEEN;
				}

				// System.out.println("" + fx + "," + fy + "  " + tx + "," + ty + "   " + moveType + "   " + baseLine);

				ChessObject objectAtTo = board.getAt(tx, ty);

				if (promoteTo != null)
					move = new Move(fx, fy, tx, ty, promoteTo);
				else
					move = new Move(fx, fy, tx, ty, moveType);
			}
			else {
				error.append("Move " + moveString + ": at from location there's no object to move");
				return null;
			}
		}
		else {
			error.append("Move " + moveString + ": format not recognized");
			return null;
		}

		for(Move currentMove : validMoves) {
			if (currentMove.getFX() == move.getFX() &&
					currentMove.getFY() == move.getFY() &&
					currentMove.getTX() == move.getTX() &&
					currentMove.getTY() == move.getTY() &&
					// FIXME: ook 'movetype' checken?
					currentMove.getPromoteTo() == move.getPromoteTo()) {
				return move;
			}
		}

		error.append("Move " + moveString + ": not in the list of valid moves");

		return null;
	}

	public static Move SANToMove(String sanMove, Board board, PlayerColor myColor, List<Move> validMoves, StringBuffer error) {
		ChessObjectType type = ChessObjectType.PAWN;
		ChessObjectType promoteTo = null;
		int baseLine = myColor == PlayerColor.WHITE ? 0 : 7;
		String from = null, to = null;
		int x = -1;
		int fx = -1, fy = -1, tx, ty;
		Move selectedMove = null;
		int garbageIndex;

		/* '+' at the end means check */
		garbageIndex = sanMove.indexOf('+');
		if (garbageIndex != -1) {
			sanMove = sanMove.substring(0, garbageIndex);
		}
		/* '#' at the end means check-mate */
		garbageIndex = sanMove.indexOf('#');
		if (garbageIndex != -1) {
			sanMove = sanMove.substring(0, garbageIndex);
		}

		/* o-o/o-o-o (castling) */
		if (sanMove.compareToIgnoreCase("o-o") == 0) {
			// FIXME: check if castling is allowed
			return new Move(4, baseLine, 6, baseLine, MoveType.CASTLING);
		}
		else if (sanMove.compareToIgnoreCase("o-o-o") == 0) {
			// FIXME: check if castling is allowed
			return new Move(4, baseLine, 2, baseLine, MoveType.CASTLING);
		}

		/* the object-type is in the first character of the string,
		 * it can be recognized by that it is an uppercase character.
		 * if it is not there (the first char is lowercase), it is
		 * a pawn
		 */
		if (isUpper(sanMove.charAt(0))) {
			type = CharToChessObjectType(sanMove.charAt(0));

			if (type == null) {
				error.append("Move " + sanMove + ": first character not recognized");
				return null;
			}

			sanMove = sanMove.substring(1);
		}

		/* if the object type is a pawn, there might be a '=Q/B/N/R'
		 * behind the string indicating the object it promotes to
		 */
		if (type == ChessObjectType.PAWN) {
			int is = sanMove.indexOf('=');

			if (is != -1) {
				promoteTo = CharToChessObjectType(sanMove.charAt(is + 1));

				if (promoteTo == null) {
					error.append("Move " + sanMove + ": promote-to character not recognized");
					return null;
				}

				sanMove = sanMove.substring(0, is);
			}
			else {
				int lastCharIndex = sanMove.length() - 1;
				char lastCharacter = sanMove.charAt(lastCharIndex);

				if (isUpper(lastCharacter)) {
					promoteTo = CharToChessObjectType(lastCharacter);

					if (promoteTo == null) {
						error.append("Move " + sanMove + ": promote-to character not recognized");
						return null;
					}

					sanMove = sanMove.substring(0, lastCharIndex);
				}
			}
		}

		/* check for capture move & split move into from/to string */
		x = sanMove.indexOf('x');
		if (x == -1) {
			x = sanMove.indexOf('X');
		}
		if (x != -1) {
			from = sanMove.substring(0, x);
			to = sanMove.substring(x + 1);
		}
		else {
			int moveLen = sanMove.length();

			if (moveLen > 2) {
				from = sanMove.substring(0, moveLen - 2);
				to = sanMove.substring(moveLen - 2);
			}
			else {
				from = null;
				to = sanMove;
			}
		}

		/* convert to-string to a position */
		tx = to.charAt(0) - 'a';
		ty = to.charAt(1) - '1';

		/* convert from-string, if available */
		if (from == null) {
			fx = fy = -1;
		}
		else if (from.length() == 2) {
			fx = from.charAt(0) - 'a';
			fy = from.charAt(1) - '1';
		}
		else if (from.length() == 1) {
			char fromChar = from.charAt(0);

			if (isDigit(fromChar)) {
				fx = -1;
				fy = fromChar - '1';
			}
			else {
				fx = fromChar - 'a';
				fy = -1;
			}
		}

		/* now find the move in the movelist */
		for(Move currentMove : validMoves) {
			if (currentMove.getTX() != tx || currentMove.getTY() != ty) {
				continue;
			}

			if (fx != -1 && currentMove.getFX() != fx) {
				continue;
			}

			if (fy != -1 && currentMove.getFY() != fy) {
				continue;
			}

			// determing the type is tricky
			if (board.getAt(currentMove.getFX(), currentMove.getFY()).getType() != type) {
				continue;
			}

			if (currentMove.getPromoteTo() != promoteTo) {
				continue;
			}

			selectedMove = currentMove;

			break;
		}

		if (selectedMove == null) {
			error.append("Move " + sanMove + ": not in list of valid moves");
			return null;
		}

		return selectedMove;
	}
}
