package com.vanheusden.pfa;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.concurrent.*;

class UciDrone extends Drone {
	UciDrone(final String host, final int port, final boolean canStop, final Map<String, String> parameters) throws IOException, InterruptedException {
		super(host, port, canStop, parameters);

		pmq = new PMQuestioner();
	}

	UciDrone(final String exe, final String path, final boolean canStop, final Map<String, String> parameters) throws IOException, InterruptedException {
		super(exe, path, canStop, parameters);

		pmq = new PMQuestioner();
	}

	int getEval() {
		return cp;
	}

	int getDepth() {
		return depth;
	}

	void initSession() throws IOException, InterruptedException {
		emit("uci");

		String line = null;

		do {
			line = input.readLine();
			if (line == null)
				break;

			if (line.length() > 8 && line.substring(0, 8).equals("id name "))
				name = line.substring(8).trim();

			if (getVerbose())
				logIgnoreIOException(getName() + " WF< " + line);
		}
		while(line != null && line.indexOf("uciok") == -1 && !Thread.interrupted());

		if (line == null) {
			logIgnoreIOException(getName() + " did not return \"uciok\"");

			disconnect();
			input = null;
		}
		else {
			emit("ucinewgame");
		}

		String ram = parameters.get("ram");
		if (ram != null)
			emit("setoption name Hash value " + ram);

		String n_threads = parameters.get("n_threads");
		if (n_threads != null)
			emit("setoption name Threads value " + n_threads);
	}

	void stopPonder() {
		try {
			// do not reset "ondering" flag here: after a stop we still
			// need to filter a result
			if (pondering.get())
				emit("stop");
		}
		catch(IOException ie) {
			logIgnoreIOException(getName() + " IO Exception in stopPonder " + ie);

			disconnect();
		}
	}

	void startPonder(final Scene sc, final PlayerColor c, final Move lastMove) {
		try {
			String cmdPosition = "position fen " + Statics.toFENString(sc.getBoard(), c, lastMove, 1, isYes(parameters.get("skipclock")));
			emit(cmdPosition);

			if (!pondering.getAndSet(true))
				emit("go ponder");
		}
		catch(IOException ie) {
			logIgnoreIOException(getName() + " IO Exception in startPonder " + ie);

			disconnect();
		}
	}

	public void startGame(final int nMoves, final long maxDuration) {
	}

	class PMQuestioner extends Drone.PMQuestioner {
		void doAsk(PMQuestion q) {
// FIXME wait until connect until deadline/2?
			try {
				logIgnoreIOException(getName() + " executing question");

				initScene.acquireUninterruptibly();
				final Scene sc = scene;
				final PlayerColor c = color;
				final int movesLeft = nr;
				long useDeadLine = deadLine - latency;
				initScene.release();

				final Move lastMove = q.getLastMove();

				startAskTs = System.currentTimeMillis();

				if (useDeadLine < startAskTs) {
					logIgnoreIOException(getName() + " no time left: skipping");
					return;
				}

				askedAnything.set(true);

				lastNps.set(0);

				//emit("ucinewgame");

				String cmdPosition = "position fen " + Statics.toFENString(sc.getBoard(), c, lastMove, 1, isYes(parameters.get("skipclock")));
				emit(cmdPosition);

				String cmdStart = "go";

				long useTimeLeft = Statics.getDelayUntil(useDeadLine);

				if (isYes(parameters.get("movetime"))) {
					cmdStart += " movetime " + useTimeLeft;
				}
				else {
					if (c == PlayerColor.WHITE)
						cmdStart += " wtime " + useTimeLeft;
					else
						cmdStart += " btime " + useTimeLeft;

					if (movesLeft == 1 && isYes(parameters.get("1movesis0moves")))
						cmdStart += " movestogo 0";
					else if (movesLeft > 0)
						cmdStart += " movestogo " + movesLeft;
				}

				//if (movesLeft > 1)
				//	cmdStart += " movestogo " + movesLeft;

				emit(cmdStart);
			}
			catch(IOException ie) {
				logIgnoreIOException(getName() + " IO Exception in startAsk " + ie);

				disconnect();
			}
		}
	}

	public void interruptThinking() {
		try {
			emit("stop");
		}
		catch(IOException ie) {
			logIgnoreIOException(getName() + " IO Exception in interruptThinking " + ie);

			disconnect();
		}
	}

	public void stopAsk() {
		askedDuration = startAskTs = askReplyTs = -1;
		depth = cp = mate = duration = nps = -1;
		nodes = -1;

		interruptThinking();
	}

	MoveResult checkInput(final String line) {
		if (line == null)
			return null;

		Statics.log("UCI: " + line);

		int bm = line.indexOf("bestmove");

		if (bm != -1 && pondering.getAndSet(false)) {
		}
		else if (bm != -1) {
			askReplyTs = System.currentTimeMillis();

			String pollMoveString = line.substring(bm);

			long took = askReplyTs - startAskTs;
			long percUsed = askedDuration > 0 ? (took * 100) / askedDuration : 100;
			logIgnoreIOException(getName() + " poll " + getName() + " choice: " + pollMoveString + ", took: " + took + "ms (" + percUsed + "%)");

			int sp = 8;

			while(sp < pollMoveString.length() && pollMoveString.charAt(sp) == ' ')
				sp++;

			String m = pollMoveString.substring(sp);

			sp = m.indexOf(" ");
			if (sp != -1)
				m = m.substring(0, sp);

			StringBuffer err = new StringBuffer("");

			initScene.acquireUninterruptibly();
			Scene sc = scene;
			PlayerColor c = color;
			initScene.release();

			if (sc == null || c == null)
				return null;

			MoveResult out = new MoveResult(cp, depth, pv, Statics.LANToMove(m, sc.getBoard(), c, sc.getMoveList(c), err), "drone:" + getName());
			depth = cp = -32767;
			pv = null;

			if (out == null)
				logIgnoreIOException(getName() + " illegal move " + m + ": " + err.toString());
			else if (out.getMove().getFX() == 0 && out.getMove().getFY() == 0 && out.getMove().getTX() == 0 && out.getMove().getTY() == 0) {
				logIgnoreIOException(getName() + " could not decide on a move yet (" + out + ")");

				// apparently a1a1 is a generic(?) way of telling that the engine
				// has no suggestion for a move. this seems to happen most when
				// the search time is relatively small (couple of seconds)
				//
				// do return the decision so that the puppetmaster does not think
				// the connection got invalid or so
				return out;
			}

			return out;
		}

		try {
			mate = duration = nps = -1;
			nodes = -1;

			int cdepth = -1, ccp = -1;
			boolean gotDepth = false, gotScore = false;
			String cpv = null;

			String [] parts = line.split(" ");

			for(int index=0; index<parts.length - 1; index++) {
				if (parts[index].equals("depth")) {
					cdepth = Integer.valueOf(parts[++index]);
					gotDepth = true;
				}
				else if (parts[index].equals("score")) {
				}
				else if (parts[index].equals("cp")) {
					ccp = Integer.valueOf(parts[++index]);
					gotScore = true;
				}
				else if (parts[index].equals("mate"))
					mate = Integer.valueOf(parts[++index]);
				else if (parts[index].equals("time"))
					duration = Integer.valueOf(parts[++index]);
				else if (parts[index].equals("nodes"))
					nodes = Long.valueOf(parts[++index]);
				else if (parts[index].equals("nps"))
					nps = Integer.valueOf(parts[++index]);
			}

			int pvIndex = line.indexOf(" pv ");
			if (pvIndex != -1)
				cpv = line.substring(pvIndex + 4);

			if (nps > 0) {
				if (getVerbose())
					logIgnoreIOException(getName() + " nps " + nps);
				lastNps.set(nps);

				if (io != null)
					io.progressCallbackNPS(nps);
			}
			else if (nodes > 0 && duration > 0) {
				nps = (int)((nodes * 1000) / duration);
				if (getVerbose())
					logIgnoreIOException(getName() + " nps " + nps);
				lastNps.set(nps);

				if (io != null)
					io.progressCallbackNPS(nps);
			}

			if (gotDepth && gotScore) {
				cp = ccp;
				depth = cdepth;
				pv = cpv;
			}

			if (io != null && depth >= 0 && duration >= 0 && nodes > 0 && pv != null)
				io.progressCallbackXboardShowThinking(depth, cp, duration / 10, nodes, pv);
		}
		catch(Exception e) { // in case of unexpected line format
			Statics.logException("info parsing", e);
		}

		return null;
	}
}
